from glob import glob
from pathlib import Path

from setuptools import setup, find_packages
import versioneer

PACKAGE_PATH = Path(__file__).parent
README = (PACKAGE_PATH / "README.md").read_text()
LICENSE = (PACKAGE_PATH / "LICENSE.txt").read_text()

setup(
    name="algotradepy",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Python Auto-Trading Framework",
    # long_description=README,
    # long_description_content_type="text/markdown",
    author="Petio Petrov",
    author_email="petioptrv@icloud.com",
    url="https://github.com/petioptrv/automated-trading",
    license=LICENSE,
    packages=find_packages(),
    scripts=glob(str(PACKAGE_PATH / "tools" / "*.py")),
    python_requires=">=3.7",  # requires setuptools>=24.2.0, pip>=9.0.0
    install_requires=[
        "numpy >=1, <2",
        "pandas >=1, <2",
        "yfinance <1",
        "pandas_datareader >=0.4.0, <1",
        "requests >=2, <3",
        "pandas_market_calendars >=1, <2",
    ],
    extras_require={
        "ibapi": ["ib_insync >=0.9, <1", "ibapi"],
        "polygon": ["websocket-client==0.57.0"],
        "dev": [
            "pytest",
            "pylint",
            "pre-commit",
            "versioneer",
            "black",
            "flake8",
            "flake8-black",
            "twine",
            "sphinx",
            "sphinx_rtd_theme",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import time
from collections import OrderedDict

import numpy as np
import pytest

from algotradepy.contracts import (
    AContract,
    StockContract,
    Exchange,
    Currency,
    ForexContract,
)
from algotradepy.orders import (
    AnOrder,
    LimitOrder,
    OrderAction,
    MarketOrder,
    TrailingStopOrder,
)
from algotradepy.trade import TradeStatus, Trade
from algotradepy.subscribable import Subscribable

AWAIT_TIME_OUT = 10


def test_acc_cash():
    pytest.importorskip("ib_insync")
    from algotradepy.brokers.ib_broker import IBBroker

    broker = IBBroker()

    acc_cash = broker.acc_cash

    broker.__del__()

    assert isinstance(acc_cash, float)
    np.testing.assert_allclose(acc_cash, 1e6, atol=1e5)


def test_datetime():
    pytest.importorskip("ib_insync")
    from algotradepy.brokers.ib_broker import IBBroker

    from datetime import datetime

    broker = IBBroker()

    broker_dt = broker.datetime
    curr_dt = datetime.now()

    broker.__del__()

    assert broker_dt.date() == curr_dt.date()
    assert broker_dt.hour == curr_dt.hour
    assert broker_dt.min == curr_dt.min
    assert np.isclose(broker_dt.second, curr_dt.second, atol=2)


def get_broker(client_id: int):
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector
    from algotradepy.brokers.ib_broker import IBBroker

    conn = build_and_start_connector(client_id=client_id)
    broker = IBBroker(ib_connector=conn)

    return broker


@pytest.fixture()
def master_broker():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import MASTER_CLIENT_ID

    broker = get_broker(client_id=MASTER_CLIENT_ID)

    yield broker

    broker.__del__()


@pytest.fixture()
def non_master_broker():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import MASTER_CLIENT_ID

    broker = get_broker(client_id=MASTER_CLIENT_ID + 1)

    yield broker

    broker.__del__()


def get_ib_test_broker(client_id: int):
    pytest.importorskip("ib_insync")
    from ib_insync.ib import IB

    class TestBroker(Subscribable, IB):
        def __init__(self):
            Subscribable.__init__(self)
            IB.__init__(self)

    tb = TestBroker()

    ip_address = "127.0.0.1"
    socket_port = 7497

    tb.connect(
        host=ip_address, port=socket_port, clientId=client_id,
    )

    return tb


@pytest.fixture()
def master_ib_test_broker():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import MASTER_CLIENT_ID

    tb = get_ib_test_broker(client_id=MASTER_CLIENT_ID)

    yield tb

    tb.reqGlobalCancel()
    tb.disconnect()


@pytest.fixture()
def non_master_ib_test_broker():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import MASTER_CLIENT_ID

    tb = get_ib_test_broker(client_id=MASTER_CLIENT_ID + 1)

    yield tb

    tb.reqGlobalCancel()
    tb.disconnect()


@pytest.fixture()
def ib_stk_contract_spy():
    pytest.importorskip("ib_insync")
    from ib_insync.contract import Stock

    contract = Stock(symbol="SPY", exchange="SMART", currency="USD")

    return contract


@pytest.fixture()
def ib_mkt_buy_order_1():
    pytest.importorskip("ib_insync")
    from ib_insync.order import MarketOrder

    order = MarketOrder(action="BUY", totalQuantity=1,)

    return order


@pytest.fixture()
def ib_mkt_sell_order_1():
    pytest.importorskip("ib_insync")
    from ib_insync.order import MarketOrder

    order = MarketOrder(action="SELL", totalQuantity=1,)

    return order


@pytest.fixture()
def ib_lmt_sell_order_2_1000():
    pytest.importorskip("ib_insync")
    from ib_insync.order import LimitOrder

    order = LimitOrder(action="SELL", totalQuantity=2, lmtPrice=1000,)

    return order


@pytest.fixture()
def spy_stock_contract():
    contract = StockContract(symbol="SPY")

    return contract


def test_subscribe_to_new_trades_non_master_raises(non_master_broker):
    def dummy_fn(*_, **__):
        pass

    with pytest.raises(AttributeError):
        master_broker.subscribe_to_new_trades(func=dummy_fn)


def test_subscribe_to_new_tws_trades(
    master_broker,
    non_master_ib_test_broker,
    ib_stk_contract_spy,
    ib_mkt_buy_order_1,
    ib_lmt_sell_order_2_1000,
):
    already_logged = []
    open_orders = OrderedDict()

    # ---- Helpers ----

    def log_new_trade(trade_: Trade):
        contract_ = trade_.contract
        order_ = trade_.order
        order_id = order_.order_id
        if order_id not in already_logged:
            open_orders[order_id] = {
                "contract": contract_,
                "order": order_,
            }
            already_logged.append(order_id)

    def await_order():
        t0 = t1 = time.time()
        while t1 - t0 <= AWAIT_TIME_OUT and len(open_orders) == 0:
            master_broker.sleep()
            t1 = time.time()

    def get_contract_and_order():
        _, open_order_dict = open_orders.popitem()
        contract_: AContract = open_order_dict["contract"]
        order_: AnOrder = open_order_dict["order"]
        return contract_, order_

    # ----------------

    master_broker.subscribe_to_new_trades(func=log_new_trade)
    non_master_ib_test_broker.placeOrder(
        ib_stk_contract_spy, ib_mkt_buy_order_1,
    )

    await_order()

    assert len(open_orders) == 1

    contract, order = get_contract_and_order()

    assert contract.symbol == "SPY"
    assert isinstance(order, MarketOrder)
    assert order.action == OrderAction.BUY
    assert order.quantity == 1

    non_master_ib_test_broker.placeOrder(
        ib_stk_contract_spy, ib_lmt_sell_order_2_1000,
    )

    await_order()

    assert len(open_orders) == 1

    contract, order = get_contract_and_order()

    assert contract.symbol == "SPY"
    assert isinstance(order, LimitOrder)
    assert order.action == OrderAction.SELL
    assert order.quantity == 2
    assert order.limit_price == 1000


def test_subscribe_to_trade_updates_non_master_raises(non_master_broker):
    def dummy_fn(*_, **__):
        pass

    with pytest.raises(AttributeError):
        master_broker.subscribe_to_trade_updates(func=dummy_fn)


def test_subscribe_to_tws_trade_updates(
    master_broker,
    non_master_ib_test_broker,
    ib_stk_contract_spy,
    ib_mkt_buy_order_1,
):
    open_orders = OrderedDict()

    def log_order_status(status_: TradeStatus):
        order_id = status_.order_id
        if order_id not in open_orders:
            open_orders[order_id] = status_

    master_broker.subscribe_to_trade_updates(func=log_order_status)
    non_master_ib_test_broker.placeOrder(
        ib_stk_contract_spy, ib_mkt_buy_order_1,
    )

    t0 = t1 = time.time()
    while t1 - t0 <= AWAIT_TIME_OUT and len(open_orders) == 0:
        master_broker.sleep()
        t1 = time.time()

    assert len(open_orders) == 1

    item = open_orders.popitem()
    status: TradeStatus = item[1]

    assert status.filled + status.remaining == 1


def request_manual_input(msg):
    from tkinter import messagebox

    messagebox.showwarning(title="Manual Input Request", message=msg)


def test_get_position_non_master_id_raises(
    non_master_broker, spy_stock_contract,
):
    with pytest.raises(AttributeError):
        non_master_broker.get_position(contract=spy_stock_contract)


def test_get_position(
    master_broker,
    spy_stock_contract,
    non_master_ib_test_broker,
    ib_stk_contract_spy,
    ib_mkt_buy_order_1,
    ib_mkt_sell_order_1,
):
    from algotradepy.connectors.ib_connector import SERVER_BUFFER_TIME

    initial_position = master_broker.get_position(contract=spy_stock_contract)

    trade = non_master_ib_test_broker.placeOrder(
        contract=ib_stk_contract_spy, order=ib_mkt_buy_order_1,
    )

    while trade.isActive() and not trade.isDone():
        non_master_ib_test_broker.sleep(SERVER_BUFFER_TIME)

    spy_position = master_broker.get_position(contract=spy_stock_contract)

    assert spy_position == initial_position + 1

    trade = non_master_ib_test_broker.placeOrder(
        contract=ib_stk_contract_spy, order=ib_mkt_sell_order_1,
    )

    while trade.isActive() and not trade.isDone():
        non_master_ib_test_broker.sleep(SERVER_BUFFER_TIME)

    spy_position = master_broker.get_position(contract=spy_stock_contract)

    assert spy_position == initial_position


def test_place_limit_order(master_ib_test_broker, non_master_broker):
    master_ib_test_broker.reqGlobalCancel()
    ib_trade = None

    def trade_receiver(trade_):
        nonlocal ib_trade
        ib_trade = trade_

    master_ib_test_broker.openOrderEvent += trade_receiver

    contract = StockContract(symbol="SPY")
    order = LimitOrder(action=OrderAction.BUY, quantity=1, limit_price=20)
    trade = Trade(contract=contract, order=order)
    placed, _ = non_master_broker.place_trade(trade=trade, await_confirm=True)

    assert isinstance(placed, bool)

    t0 = time.time()
    while ib_trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        non_master_broker.sleep()

    assert ib_trade.contract.symbol == "SPY"
    assert ib_trade.order.orderType == "LMT"
    assert ib_trade.order.totalQuantity == 1
    assert ib_trade.order.action == "BUY"
    assert ib_trade.order.lmtPrice == 20


def test_receive_limit_order(non_master_ib_test_broker, master_broker):
    from ib_insync.order import LimitOrder as IBLimitOrder
    from ib_insync.contract import Stock as IBStock

    trade: Trade = None

    def trade_receiver(trade_):
        nonlocal trade
        trade = trade_

    master_broker.subscribe_to_new_trades(func=trade_receiver)

    contract = IBStock(symbol="SPY", exchange="SMART", currency="USD")
    order = IBLimitOrder(action="BUY", totalQuantity=1, lmtPrice=10)
    non_master_ib_test_broker.placeOrder(contract=contract, order=order)

    t0 = time.time()
    while trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        master_broker.sleep()

    assert trade.contract.symbol == "SPY"
    assert isinstance(trade.order, LimitOrder)
    assert trade.order.quantity == 1
    assert trade.order.action == OrderAction.BUY
    assert trade.order.limit_price == 10


@pytest.mark.parametrize("aux_price,trailing_percent", [(5, None), (None, 5)])
def test_place_trailing_stop_order(
    aux_price, trailing_percent, master_ib_test_broker, non_master_broker,
):
    from ib_insync.util import UNSET_DOUBLE

    master_ib_test_broker.reqGlobalCancel()
    ib_trade = None

    def trade_receiver(trade_):
        nonlocal ib_trade
        ib_trade = trade_

    master_ib_test_broker.openOrderEvent += trade_receiver

    contract = StockContract(symbol="SPY")
    order = TrailingStopOrder(
        action=OrderAction.BUY,
        quantity=1,
        trail_stop_price=10,
        aux_price=aux_price,
        trail_percent=trailing_percent,
    )
    trade = Trade(contract=contract, order=order)
    placed, _ = non_master_broker.place_trade(trade=trade, await_confirm=True)

    assert isinstance(placed, bool)

    t0 = time.time()
    while ib_trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        non_master_broker.sleep()

    assert ib_trade.contract.symbol == "SPY"
    assert ib_trade.order.orderType == "TRAIL"
    assert ib_trade.order.totalQuantity == 1
    assert ib_trade.order.action == "BUY"
    assert ib_trade.order.trailStopPrice == 10
    if aux_price is not None:
        assert ib_trade.order.auxPrice == aux_price
    else:
        assert ib_trade.order.auxPrice == UNSET_DOUBLE
    if trailing_percent is not None:
        assert ib_trade.order.trailingPercent == trailing_percent
    else:
        assert ib_trade.order.trailingPercent == UNSET_DOUBLE


@pytest.mark.parametrize("aux_price,trailing_percent", [(5, None), (None, 5)])
def test_receive_trailing_stop_order(
    aux_price, trailing_percent, non_master_ib_test_broker, master_broker,
):
    from ib_insync.util import UNSET_DOUBLE
    from ib_insync.order import Order as IBorder
    from ib_insync.contract import Stock as IBStock

    trade: Trade = None

    def trade_receiver(trade_):
        nonlocal trade
        trade = trade_

    master_broker.subscribe_to_new_trades(func=trade_receiver)

    contract = IBStock(symbol="SPY", exchange="SMART", currency="USD")
    order = IBorder(
        orderType="TRAIL",
        action="BUY",
        totalQuantity=1,
        trailStopPrice=10,
        auxPrice=aux_price or UNSET_DOUBLE,
        trailingPercent=trailing_percent or UNSET_DOUBLE,
    )
    non_master_ib_test_broker.placeOrder(contract=contract, order=order)

    t0 = time.time()
    while trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        master_broker.sleep()

    assert trade.contract.symbol == "SPY"
    assert isinstance(trade.order, TrailingStopOrder)
    assert trade.order.quantity == 1
    assert trade.order.action == OrderAction.BUY
    assert trade.order.trail_stop_price == 10
    assert trade.order.aux_price == aux_price
    assert trade.order.trail_percent == trailing_percent


def test_place_forex_order(master_ib_test_broker, non_master_broker):
    master_ib_test_broker.reqGlobalCancel()
    ib_trade = None

    def trade_receiver(trade_):
        nonlocal ib_trade
        ib_trade = trade_

    master_ib_test_broker.openOrderEvent += trade_receiver

    contract = ForexContract(symbol="EUR", currency=Currency.USD)
    order = LimitOrder(action=OrderAction.BUY, quantity=20000, limit_price=1)
    trade = Trade(contract=contract, order=order)
    placed, _ = non_master_broker.place_trade(trade=trade, await_confirm=True)

    assert isinstance(placed, bool)

    t0 = time.time()
    while ib_trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        non_master_broker.sleep()

    assert ib_trade.contract.symbol == "EUR"
    assert ib_trade.contract.secType == "CASH"
    assert ib_trade.order.orderType == "LMT"
    assert ib_trade.order.totalQuantity == 20000
    assert ib_trade.order.action == "BUY"
    assert ib_trade.order.lmtPrice == 1


def test_receive_forex_order(non_master_ib_test_broker, master_broker):
    from ib_insync.order import LimitOrder as IBLimitOrder
    from ib_insync.contract import Forex as IBForex

    trade: Trade = None

    def trade_receiver(trade_):
        nonlocal trade
        trade = trade_

    master_broker.subscribe_to_new_trades(func=trade_receiver)

    contract = IBForex(pair="EURUSD")
    order = IBLimitOrder(action="BUY", totalQuantity=20000, lmtPrice=1)
    non_master_ib_test_broker.placeOrder(contract=contract, order=order)

    t0 = time.time()
    while trade is None and time.time() - t0 <= AWAIT_TIME_OUT:
        master_broker.sleep()

    assert trade.contract.symbol == "EUR"
    assert isinstance(trade.contract, ForexContract)
    assert trade.order.quantity == 20000
    assert trade.order.action == OrderAction.BUY
    assert trade.order.limit_price == 1


@pytest.mark.parametrize(
    "exchange,currency,symbol,size",
    [
        (Exchange.NYSE, Currency.USD, "SPY", 1),
        (Exchange.NASDAQ, Currency.USD, "TSLA", 1),
        (Exchange.AMEX, Currency.USD, "TSLA", 1),
        (Exchange.ARCA, Currency.USD, "TSLA", 1),
        (Exchange.TSE, Currency.CAD, "AQN", 1),
        (Exchange.VENTURE, Currency.CAD, "VTI", 1),
        (Exchange.FWB, Currency.EUR, "FME", 1),
        (Exchange.IBIS, Currency.EUR, "SAP", 1),
        (Exchange.VSE, Currency.EUR, "SBO", 1),
        (Exchange.LSE, Currency.GBP, "BRBY", 1),
        (Exchange.BATEUK, Currency.GBP, "GENL", 1),
        (Exchange.ENEXT_BE, Currency.EUR, "BAR", 1),
        (Exchange.SBF, Currency.EUR, "BRE", 1),
        (Exchange.AEB, Currency.EUR, "LQDA", 1),
        (Exchange.SEHK, Currency.HKD, "98", 1000),
        (Exchange.ASX, Currency.AUD, "CML", 1),
        (Exchange.TSEJ, Currency.JPY, "2334", 100),
    ],
)
def test_exchanges_and_currencies(
    exchange, currency, symbol, size, non_master_broker, master_ib_test_broker,
):
    time.sleep(1)
    contract = StockContract(
        symbol=symbol, exchange=exchange, currency=currency,
    )
    order = LimitOrder(action=OrderAction.BUY, quantity=size, limit_price=10)
    trade = Trade(contract=contract, order=order)
    placed, _ = non_master_broker.place_trade(trade=trade, await_confirm=True)

    assert isinstance(placed, bool)



from algotradepy.brokers.sim_broker import SimulationBroker

__all__ = ["SimulationBroker"]

try:
    from algotradepy.brokers.ib_broker import IBBroker

    __all__.append("IBBroker")
except ImportError:
    pass



from enum import Enum
from typing import Optional

from algotradepy.contracts import AContract
from algotradepy.orders import AnOrder
from algotradepy.utils import ReprAble, Comparable


class TradeState(Enum):
    """The possible trade states.

    Values
    ------
    * SUBMITTED
    * FILLED
    * CANCELLED
    * PENDING
    * INACTIVE
    """

    SUBMITTED = "SUBMITTED"
    FILLED = "FILLED"
    CANCELLED = "CANCELLED"
    PENDING = "PENDING"
    INACTIVE = "INACTIVE"


class TradeStatus(ReprAble, Comparable):
    """Defines the status of an order.

    Parameters
    ----------
    state : ~algotradepy.trade.TradeState
        The state of the order.
    filled : float
        How much of the order was filled. Cannot exceed the order size.
    remaining : float
        How much of the order is remaining.
    ave_fill_price : float
        The average price at which the order was filled.
    order_id : int, optional, default None
        The associated order's ID.
    """

    def __init__(
        self,
        state: TradeState,
        filled: float,
        remaining: float,
        ave_fill_price: float,
        order_id: Optional[int] = None,
    ):
        super().__init__()
        self._order_id = order_id
        self._state = state
        self._filled = filled
        self._remaining = remaining
        self._ave_fill_price = ave_fill_price

    @property
    def order_id(self) -> int:
        return self._order_id

    @property
    def state(self):
        return self._state

    @property
    def filled(self):
        return self._filled

    @property
    def remaining(self):
        return self._remaining

    @property
    def ave_fill_price(self):
        return self._ave_fill_price


class Trade(ReprAble):
    """This class defines a trade.

    Parameters
    ----------
    contract : ~algotradepy.contracts.AContract
        The contract definition for this trade.
    order : ~algotradepy.orders.AnOrder
        The order definition for this trade.
    status : ~algotradepy.trade.TradeStatus, optional, default None
        The status of the trade.
    """

    def __init__(
        self,
        contract: AContract,
        order: AnOrder,
        status: Optional[TradeStatus] = None,
    ):
        super().__init__()
        self._contract = contract
        self._order = order
        self._status = status

    @property
    def contract(self) -> AContract:
        return self._contract

    @property
    def order(self) -> AnOrder:
        return self._order

    @property
    def status(self) -> TradeStatus:
        return self._status

    @status.setter
    def status(self, new_status: TradeStatus):
        self._status = new_status

    def __hash__(self):
        h = hash((hash(self._contract), hash(self._order)),)
        return h

    def __eq__(self, other) -> bool:
        equal = True

        if not isinstance(other, Trade):
            equal = False
        elif self.contract != other.contract or self.order != other.order:
            equal = False

        return equal



from typing import Optional

try:
    from ib_insync import IB
except ImportError as e:
    raise ImportError(
        f"Original Error: {e}"
        "\nThe IB API is not installed. Please reinstall using"
        " 'pip install algotradepy[ibapi]' or manually"
        " from https://www.interactivebrokers.com/en/index.php?f=5041."
    )

from algotradepy.subscribable import Subscribable

MASTER_CLIENT_ID = 0
_NEXT_VALID_CLIENT_ID = -1
SERVER_BUFFER_TIME = 0.1


def _get_client_id() -> int:
    global _NEXT_VALID_CLIENT_ID
    _NEXT_VALID_CLIENT_ID += 1
    return _NEXT_VALID_CLIENT_ID


class IBConnector(IB):
    """A wrapper around `ib_insync.IB`."""

    def __init__(self):
        Subscribable.__init__(self)
        IB.__init__(self)
        self._client_id = None

    @property
    def client_id(self) -> Optional[int]:
        return self._client_id

    def connect(
        self,
        host: str = "127.0.0.1",
        port: int = 7497,
        client_id: int = 1,
        timeout: float = 4,
        readonly: bool = False,
        account: str = "",
    ):
        IB.connect(
            self=self,
            host=host,
            port=port,
            clientId=client_id,
            timeout=timeout,
            readonly=readonly,
            account=account,
        )
        self._client_id = client_id


def build_and_start_connector(
    receiver: str = "workstation",
    trading_mode: str = "paper",
    socket_port: Optional[int] = None,
    client_id: Optional[int] = None,
) -> IBConnector:
    """Builds and prepares a connector for use.

    The function takes care of thread instantiation and ensure that the
    communication with the server is responsive before returning the
    instantiated connector object.

    Parameters
    ----------
    receiver : str, default "workstation"
        See IBConnector documentation for details.
    trading_mode : str, default "paper"
        See IBConnector documentation for details.
    socket_port : int, optional, default None
        See IBConnector documentation for details.
    client_id : int, optional, default None
        See IBConnector documentation for details.

    Returns
    -------
    connector : IBConnector
        The instantiated connector, ready for use.

    Notes
    -----
    Socket Port Mapping:
    +---------------+----------------+-------------+
    | `receiver`    | `trading_mode` | Socket Port |
    +---------------+----------------+-------------+
    | "workstation" | "live"         | 7496        |
    +---------------+----------------+-------------+
    | "workstation" | "paper"        | 7497        |
    +---------------+----------------+-------------+
    | "gateway"     | "live"         | 4001        |
    +---------------+----------------+-------------+
    | "gateway"     | "paper"        | 4002        |
    +---------------+----------------+-------------+

    """

    ip_address = "127.0.0.1"

    if socket_port is not None:
        socket_port = socket_port
    elif receiver == "workstation" and trading_mode == "live":
        socket_port = 7496
    elif receiver == "workstation" and trading_mode == "paper":
        socket_port = 7497
    elif receiver == "gateway" and trading_mode == "live":
        socket_port = 4001
    elif receiver == "gateway" and trading_mode == "paper":
        socket_port = 4002
    else:
        raise ValueError(
            f"Unrecognized socket port configuration receiver={receiver}"
            f" trading_mode={trading_mode}. Please consult the"
            f" documentation for valid values."
        )

    if client_id is None:
        client_id = _get_client_id()

    connector = IBConnector()
    connector.connect(host=ip_address, port=socket_port, client_id=client_id)

    return connector



from datetime import timedelta, date

from algotradepy.contracts import StockContract
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.historical.providers.yahoo_provider import (
    YahooHistoricalProvider,
)

if __name__ == "__main__":
    provider = YahooHistoricalProvider()
    hr = HistoricalRetriever(provider=provider)
    end_date = date.today()
    hr.retrieve_bar_data(
        contract=StockContract(symbol="SPY"),
        bar_size=timedelta(days=1),
        start_date=date(2000, 1, 1),
        end_date=end_date,
    )



from datetime import datetime
from typing import Optional, Tuple, Dict, List, Callable

from algotradepy.ib_utils import IBBase

try:
    import ib_insync
except ImportError:
    raise ImportError(
        "Optional package ib_insync not install. Please install"
        " using 'pip install ib_insync'."
    )
from ibapi.account_summary_tags import (
    AccountSummaryTags as _AccountSummaryTags,
)
from ib_insync.order import Trade as _IBTrade

from algotradepy.brokers.base import ABroker
from algotradepy.connectors.ib_connector import (
    IBConnector,
    MASTER_CLIENT_ID,
    SERVER_BUFFER_TIME,
)
from algotradepy.contracts import (
    AContract,
    are_loosely_equal_contracts,
)
from algotradepy.trade import Trade


class IBBroker(IBBase, ABroker):
    """Interactive Brokers Broker class.

    This class implements the ABroker interface for use with the Interactive
    Brokers API. It is an adaptor to the IBConnector, exposing the relevant
    methods.

    Parameters
    ----------
    simulation : bool, default True
        Ignored if parameter `ib_connector` is supplied.
    ib_connector : IBConnector, optional, default None
        A custom instance of the `IBConnector` can be supplied. If not provided,
        it is assumed that the receiver is TWS (see `IBConnector`'s
        documentation for more details).
    """

    def __init__(
        self,
        simulation: bool = True,
        ib_connector: Optional[IBConnector] = None,
    ):
        ABroker.__init__(self, simulation=simulation)
        IBBase.__init__(self, simulation=simulation, ib_connector=ib_connector)

    @property
    def acc_cash(self) -> float:
        acc_summary = self._ib_conn.accountSummary()
        acc_values = []

        for s in acc_summary:
            if s.tag == _AccountSummaryTags.TotalCashValue:
                acc_values.append(float(s.value))

        acc_value = sum(acc_values)

        return acc_value

    @property
    def datetime(self) -> datetime:
        dt = self._ib_conn.reqCurrentTime()
        dt = dt.astimezone()  # localize the timezone
        return dt

    @property
    def trades(self) -> List[Trade]:
        ib_trades = self._ib_conn.trades()
        trades = [
            self._from_ib_trade(ib_trade=ib_trade) for ib_trade in ib_trades
        ]
        return trades

    @property
    def open_trades(self) -> List[Trade]:
        open_ib_trades = self._ib_conn.openTrades()
        open_trades = [
            self._from_ib_trade(ib_trade=ib_trade)
            for ib_trade in open_ib_trades
        ]
        return open_trades

    def subscribe_to_new_trades(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        if fn_kwargs is None:
            fn_kwargs = {}

        def submitted_order_filter(ib_trade: _IBTrade):
            trade = self._from_ib_trade(ib_trade=ib_trade)
            func(trade, **fn_kwargs)

        self._ib_conn.openOrderEvent += submitted_order_filter

    def subscribe_to_trade_updates(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        if fn_kwargs is None:
            fn_kwargs = {}

        def order_status_filter(ib_trade: _IBTrade):
            status = self._from_ib_status(ib_order_status=ib_trade.orderStatus)
            func(status, **fn_kwargs)

        self._ib_conn.orderStatusEvent += order_status_filter

    def place_trade(
        self, trade: Trade, *args, await_confirm: bool = False,
    ) -> Tuple[bool, Trade]:
        ib_contract = self._to_ib_contract(contract=trade.contract)
        ib_order = self._to_ib_order(order=trade.order)

        ib_trade = self._ib_conn.placeOrder(
            contract=ib_contract, order=ib_order,
        )

        placed = True
        confirmed = False
        while await_confirm and not confirmed:
            status = ib_trade.orderStatus.status
            if "Submitted" in status or status == "Filled":
                placed = True
                confirmed = True
            elif "Cancelled" in status:
                placed = False
                confirmed = True
            self.sleep(SERVER_BUFFER_TIME)

        trade = self._from_ib_trade(ib_trade=ib_trade)

        return placed, trade

    def cancel_trade(self, trade: Trade):
        ib_order = self._to_ib_order(order=trade.order)
        self._ib_conn.cancelOrder(order=ib_order)

    def get_position(
        self,
        contract: AContract,
        *args,
        account: Optional[str] = None,
        **kwargs,
    ) -> float:
        if self._ib_conn.client_id != MASTER_CLIENT_ID:
            raise AttributeError(
                f"This client ID cannot request positions. Please use a broker"
                f" instantiated with the master client ID ({MASTER_CLIENT_ID})"
                f" to request positions."
            )

        if account is None:
            account = ""

        pos = 0
        positions = self._ib_conn.positions(account=account)
        for position in positions:
            pos_contract = self._from_ib_contract(
                ib_contract=position.contract
            )
            if are_loosely_equal_contracts(
                loose=contract, well_defined=pos_contract,
            ):
                pos = position.position
                break

        return pos

    def get_transaction_fee(self) -> float:
        # TODO: implement
        raise NotImplementedError



from abc import ABC
from enum import Enum
from typing import Optional, List

from algotradepy.order_conditions import ACondition
from algotradepy.utils import ReprAble


class OrderAction(Enum):
    """The available order actions.

    Values
    ------
    * BUY
    * SELL
    """

    BUY = "BUY"
    SELL = "SELL"


class TimeInForce(Enum):
    """The order time-in-force options.

    Values
    ------
    * DAY
    * GTC
    """

    DAY = "DAY"
    GTC = "GTC"


class AnOrder(ABC, ReprAble):
    """The abstract order class defining the basic order properties.

    Parameters
    ----------
    action : ~algotradepy.orders.OrderAction
        The order action.
    quantity : float
        The order size.
    order_id : int, optional, default None
        The order ID.
    time_in_force : ~algotradepy.orders.TimeInForce, optional, default None
        Time-in-force option.
    conditions : list of ~algotradepy.orders.ACondition, optional, default None
        A list of conditions attached to the order.
    parent_id : int, optional, default None
        The parent order's ID.
    """

    def __init__(
        self,
        action: OrderAction,
        quantity: float,
        order_id: Optional[int] = None,
        time_in_force: Optional[TimeInForce] = None,
        conditions: Optional[List[ACondition]] = None,
        parent_id: Optional[int] = None,
    ):
        super().__init__()
        self._order_id = order_id
        self._action = action
        self._quantity = quantity
        self._time_in_force = time_in_force
        if conditions is None:
            conditions = []
        self._conditions = conditions
        self._parent_id = parent_id

    @property
    def order_id(self) -> int:
        return self._order_id

    @property
    def action(self) -> OrderAction:
        return self._action

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def time_in_force(self) -> TimeInForce:
        return self._time_in_force

    @property
    def conditions(self) -> List[ACondition]:
        return self._conditions

    @property
    def parent_id(self):
        return self._parent_id


class MarketOrder(AnOrder):
    """A market order definition."""

    def __init__(
        self,
        action: OrderAction,
        quantity: float,
        order_id: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(
            order_id=order_id, action=action, quantity=quantity, **kwargs
        )


class LimitOrder(AnOrder):
    """A limit order definition.

    Parameters
    ----------
    action
    quantity
    limit_price : float
        The limit price for this order.
    order_id
    kwargs
    """

    def __init__(
        self,
        action: OrderAction,
        quantity: float,
        limit_price: float,
        order_id: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(
            order_id=order_id, action=action, quantity=quantity, **kwargs
        )
        self._limit_price = limit_price

    @property
    def limit_price(self) -> float:
        return self._limit_price


class TrailingStopOrder(AnOrder):
    """A trailing stop order definition.

    Parameters
    ----------
    action
    quantity
    trail_stop_price : float, optional, default None
        The Trailing stop price for this order.
    aux_price : float, optional, default None
        Can be specified instead of `trail_percent`, in which case the order is
        executed at `aux_price` away from `trailing_stop_price`. Must be
        provided if `trail_percent` is left blank.
    trail_percent : float, optional, default None
        The percentage offset from `trail_stop_price` at which to execute.
        Must be provided if `aux_price` is not set.
    order_id
    kwargs
    """

    def __init__(
        self,
        action: OrderAction,
        quantity: float,
        trail_stop_price: Optional[float] = None,
        aux_price: Optional[float] = None,
        trail_percent: Optional[float] = None,
        order_id: Optional[int] = None,
        **kwargs,
    ):
        self._validate(
            aux_price=aux_price, trail_percent=trail_percent,
        )
        super().__init__(
            order_id=order_id, action=action, quantity=quantity, **kwargs
        )
        self._trail_stop_price = trail_stop_price
        self._trail_percent = trail_percent
        self._aux_price = aux_price

    @property
    def trail_stop_price(self) -> Optional[float]:
        return self._trail_stop_price

    @property
    def trail_percent(self) -> Optional[float]:
        return self._trail_percent

    @property
    def aux_price(self):
        return self._aux_price

    @staticmethod
    def _validate(
        aux_price: Optional[float], trail_percent: Optional[float],
    ):
        if (trail_percent is None) == (aux_price is None):
            raise ValueError(
                f"Exactly one of aux_price or trail_percent must be specified"
                f" for {TrailingStopOrder.__name__}."
            )
